package com.contentsquare.flutter

import android.app.Activity
import android.content.Context
import android.src.main.kotlin.com.contentsquare.flutter.infrastructure.core.states.AppInfoProvider
import android.view.View
import android.view.ViewGroup
import androidx.annotation.NonNull
import androidx.lifecycle.DefaultLifecycleObserver
import androidx.lifecycle.LifecycleOwner
import io.flutter.embedding.engine.plugins.FlutterPlugin
import io.flutter.embedding.engine.plugins.activity.ActivityAware
import io.flutter.embedding.engine.plugins.activity.ActivityPluginBinding
import io.flutter.plugin.common.MethodCall
import io.flutter.plugin.common.MethodChannel
import io.flutter.plugin.common.MethodChannel.MethodCallHandler
import io.flutter.plugin.common.MethodChannel.Result
import com.contentsquare.android.api.bridge.flutter.ExternalViewGraphListener
import com.contentsquare.android.api.bridge.flutter.ExternalViewGraphResult
import com.contentsquare.android.api.bridge.flutter.FlutterInterface
import com.contentsquare.android.api.bridge.xpf.ExternalBridgeInterface
import com.contentsquare.android.api.bridge.xpf.ExternalBridgeType
import com.contentsquare.flutter.infrastructure.bridges.FlutterInterfaceBridge
import com.contentsquare.flutter.infrastructure.bridges.XpfInterfaceBridge
import com.contentsquare.flutter.infrastructure.core.TargetViewPathCreator
import com.contentsquare.flutter.infrastructure.core.method_call_executor.MethodCallExecutor
import com.contentsquare.flutter.infrastructure.core.method_call_executor.agents.MethodCallAgent
import com.contentsquare.flutter.infrastructure.core.method_call_executor.data.ExecutionData
import com.contentsquare.flutter.infrastructure.core.platform_views.FlutterEngineProvider
import com.contentsquare.flutter.infrastructure.core.platform_views.WebViewRegistrationManager
import com.contentsquare.flutter.infrastructure.core.states.ContentsquarePluginState
import com.contentsquare.flutter.infrastructure.core.states.StateObject
import io.flutter.Log
import io.flutter.embedding.android.FlutterImageView
import io.flutter.embedding.android.FlutterSurfaceView
import io.flutter.embedding.android.FlutterTextureView
import io.flutter.embedding.android.FlutterView
import io.flutter.embedding.engine.FlutterEngine
import io.flutter.embedding.engine.mutatorsstack.FlutterMutatorView
import io.flutter.embedding.engine.plugins.lifecycle.HiddenLifecycleReference
import java.lang.ref.WeakReference

internal object ViewFinder {

    fun findFlutterView(activity: Activity): FlutterView? {
        val view = activity.window.decorView
        return findRecursively(view, FlutterView::class.java)
    }

    fun findFlutterSurfaceView(flutterView: FlutterView): FlutterSurfaceView? {
        return findRecursively(flutterView, FlutterSurfaceView::class.java)
    }

    fun findFlutterTextureView(flutterView: FlutterView): FlutterTextureView? {
        return findRecursively(flutterView, FlutterTextureView::class.java)
    }

    private fun <T> findRecursively(view: View, clazz: Class<T>): T? {
        if (clazz.isInstance(view)) {
            @Suppress("UNCHECKED_CAST")
            return view as T
        }
        if (view is ViewGroup) {
            for (childIndex in 0 until view.childCount) {
                val viewFound = findRecursively(view.getChildAt(childIndex), clazz)
                if (viewFound != null) {
                    return viewFound
                }
            }
        }
        return null
    }
}

/** ContentsquarePlugin */
class ContentsquarePlugin : FlutterPlugin, ActivityAware, MethodCallHandler,
    DefaultLifecycleObserver, FlutterEngineProvider {
    private val targetViewPathCreator = TargetViewPathCreator()

    private var flutterEngineReference: WeakReference<FlutterEngine>? = null
    private lateinit var channel: MethodChannel
    private var context: Context? = null
    private lateinit var view: View
    private lateinit var activity: Activity
    private lateinit var executor: MethodCallExecutor
    private var featuresState = ContentsquarePluginState
    private var flutterInterfaceBridge = FlutterInterfaceBridge.default()
    private val xpfInterfaceBridge = XpfInterfaceBridge.default()
    private val externalBridgeInterface = object : ExternalBridgeInterface {
        override val bridgeType: ExternalBridgeType = ExternalBridgeType.FLUTTER

        override fun notifyApiErrorsEnabled(enable: Boolean) {
            featuresState.setApiErrorsState(StateObject(enable))
            channel.invokeMethod(
                ChannelConst.Method.REQUEST_ENABLE_API_ERROR,
                mapOf(ChannelConst.Key.ENABLE_API_ERRORS to enable)
            )
        }

        // TODO: change this method to accept the view id as a parameter when the Native SDK will be updated in FLUT-1215
        override fun notifySessionReplayEnabled(enable: Boolean) {
            val containerViewId: Int? = null
            var pixelRatio = 1.0f
            if (enable) {
                pixelRatio = context?.resources?.displayMetrics?.density ?: 1.0f
            }

            featuresState.setContainerViewIdState(StateObject(containerViewId))
            featuresState.setSessionReplayState(StateObject(enable))
            featuresState.setPixelRatioState(StateObject(pixelRatio.toDouble()))

            channel.invokeMethod(
                ChannelConst.Method.REQUEST_ENABLE_SR,
                mapOf(ChannelConst.Key.ENABLE_SR to enable,
                        ChannelConst.Key.CONTAINER_VIEW_ID to containerViewId,
                        ChannelConst.Key.PIXEL_RATIO to pixelRatio)
            )
        }


        override fun notifyCrashReportingEnabled(enable: Boolean) {
            featuresState.setCrashReporterState(StateObject(enable))
            channel.invokeMethod(
                ChannelConst.Method.REQUEST_ENABLE_CRASH_REPORTER,
                mapOf(ChannelConst.Key.ENABLE_CRASH_REPORTER to enable)
            )
        }

        override fun notifyFeatureFlagsEnabled(featureFlags: List<Map<String, Any>>) {
            featuresState.setFeatureFlagsState(StateObject(featureFlags))

            channel.invokeMethod(
                ChannelConst.Method.UPDATE_FEATURE_FLAGS,
                mapOf(ChannelConst.Key.FEATURE_FLAGS to featureFlags)
            )
        }

        override fun setTagId(tagId: String) {
            Log.d("CSLIB", "setTagId not yet implemented")
        }

        override fun notifyCsInAppEnabled(enabled: Boolean) {
            Log.d("CSLIB", "CS InApp feature state changes " + enabled.toString())
            featuresState.setCsInAppState(StateObject(enabled))

            channel.invokeMethod(
                ChannelConst.Method.REQUEST_ENABLE_CS_IN_APP,
                mapOf(ChannelConst.Key.ENABLE_CS_IN_APP to enabled)
            )
        }
    }

    private fun registerNewExternalViewFromActivity(@NonNull activity: Activity) {
        view = activity.findViewById(android.R.id.content) as View


        val flutterView = getFlutterViewIfAvailable() ?: return
        flutterEngineReference = WeakReference(flutterView.attachedFlutterEngine)
        val renderView = ViewFinder.findFlutterSurfaceView(flutterView)
                ?: ViewFinder.findFlutterTextureView(flutterView)
                ?: return

        FlutterInterface.excludeExternalView(
            FlutterImageView::class.java,
            FlutterMutatorView::class.java,
        );

        // We register this `renderView` that can be either a [FlutterSurfaceView] or
        // a [FlutterTextureView]. Flutter use one of those to render the Flutter UI
        // based on the [RenderMode] used.
        // [RenderMode.surface] will add [FlutterSurfaceView] and Flutter will use it and
        // it is the most common/recommended render mode.
        // [RenderMode.texture] will add [FlutterTextureView] and Flutter will use it and
        // it is used when you want to use Flutter in between others native views.
        // We don't register the [FlutterView] directly cause inside it there's some subviews
        // that the native SDK should handle like registered [WebView]s.
        flutterInterfaceBridge.registerExternalView(renderView,
            object : ExternalViewGraphListener {
                override fun takeSnapShot(
                    view: View, path: String, result: ExternalViewGraphResult
                ) {
                    val externalViewGraphResult = result

                    val targetViewPath: String = targetViewPathCreator.create(renderView, path)

                    val pixelRatio: Double =
                        context?.resources?.displayMetrics?.density?.toDouble() ?: 1.0

                    channel.invokeMethod(
                        ChannelConst.Method.REQUEST_FLUTTER_SNAPSHOT,
                        mapOf(
                            ChannelConst.Key.ROOT_PATH to targetViewPath,
                            ChannelConst.Key.PIXEL_RATIO to pixelRatio
                        ),
                        object : Result {
                            override fun success(result: Any?) {
                                externalViewGraphResult.onSnapshotTaken(
                                    view, result as String
                                )
                            }

                            override fun error(
                                errorCode: String, errorMessage: String?, errorDetails: Any?
                            ) {
                                Log.d("CSLIB", "$errorCode, $errorMessage, $errorDetails")
                            }

                            override fun notImplemented() {
                                Log.d("CSLIB", "notImplemented")
                            }
                        })
                }
            })

        xpfInterfaceBridge.registerExternalBridge(externalBridgeInterface)

    }

    private fun getFlutterViewIfAvailable(): FlutterView? {
        val flutterView = ViewFinder.findFlutterView(activity)
        if (flutterView == null) {
            Log.e("CSLIB", "ContentsquarePlugin, no Flutter view found.")
            return null
        }
        return flutterView
    }

    private fun setAppInfo(context: Context) {
        val appInfo = AppInfoProvider.get(context)
        featuresState.setAppInfoState(StateObject(appInfo))
    }

    override fun onAttachedToEngine(
        @NonNull flutterPluginBinding: FlutterPlugin.FlutterPluginBinding
    ) {
        channel = MethodChannel(flutterPluginBinding.binaryMessenger, "contentsquare")
        channel.setMethodCallHandler(this)
        context = flutterPluginBinding.applicationContext

        executor = MethodCallExecutor()
        setAppInfo(flutterPluginBinding.applicationContext)
        WebViewRegistrationManager.initInstance(this)
    }

    override fun onAttachedToActivity(@NonNull activityPluginBinding: ActivityPluginBinding) {
        (activityPluginBinding.lifecycle as HiddenLifecycleReference).lifecycle.addObserver(this)
        activity = activityPluginBinding.activity
    }

    override fun onResume(owner: LifecycleOwner) {
        registerNewExternalViewFromActivity(activity)
    }

    override fun onReattachedToActivityForConfigChanges(@NonNull activityPluginBinding: ActivityPluginBinding) {
        registerNewExternalViewFromActivity(activityPluginBinding.activity)
    }

    override fun onMethodCall(@NonNull call: MethodCall, @NonNull result: Result) {
        val agent = MethodCallAgent.of(call.method, channel=channel)
        val data = ExecutionData(call = call, context = context)
        executor.execute(agent = agent, data = data, result = result)
    }

    override fun onDetachedFromEngine(@NonNull flutterPluginBinding: FlutterPlugin.FlutterPluginBinding) {
        channel.setMethodCallHandler(null)
        context = null
    }

    override fun onDetachedFromActivityForConfigChanges() {
        flutterInterfaceBridge.unRegisterExternalView(view)
        xpfInterfaceBridge.unregisterExternalBridge(externalBridgeInterface)
    }

    override fun onDetachedFromActivity() {
        flutterInterfaceBridge.unRegisterExternalView(view)
        xpfInterfaceBridge.unregisterExternalBridge(externalBridgeInterface)
    }

    override fun getFlutterEngine(): FlutterEngine? {
        return flutterEngineReference?.get()
    }
}
