package com.contentsquare.flutter.infrastructure.core.method_call_executor.agents

import com.contentsquare.flutter.infrastructure.core.method_call_executor.data.ExecutionData
import com.contentsquare.flutter.infrastructure.core.method_call_executor.responses.ExecutionResponse
import com.contentsquare.flutter.infrastructure.core.method_call_executor.responses.FailedExecutionResponse
import com.contentsquare.flutter.infrastructure.core.method_call_executor.responses.FailedExecutionResponseReason
import io.flutter.plugin.common.MethodCall
import io.flutter.plugin.common.MethodChannel
import kotlin.reflect.KClass


/**
 * Agent used to execute specific method from [MethodCall].
 */
internal abstract class MethodCallAgent(private val nativeExceptionMessage: String?) {
    /**
     * Process the execution of the method
     *
     * @param data Data used during execution
     */
    protected abstract fun execute(data: ExecutionData): ExecutionResponse

    /**
     * Process the execution of the method and catch any error/exception.
     *
     * @param data Data used during execution
     */
    fun safeExecute(data: ExecutionData): ExecutionResponse {
        return try {
            execute(data)
        } catch (t: Throwable) {
            FailedExecutionResponse(
                FailedExecutionResponseReason.NATIVE_EXCEPTION,
                nativeExceptionMessage,
                t
            )
        }
    }

    companion object Factory {
        /**
         * Creates a new instance implementation of [MethodCallAgent].
         *
         * @param method the method to execute used to create the specific [MethodCallAgent]
         */
        fun of(method: String, channel: MethodChannel): MethodCallAgent {
            val agentMethod = AgentMethod.values().firstOrNull { agentMethod ->
                agentMethod.method == method
            } ?: return MissingMethodCallAgent(method = method)

            return when (agentMethod) {
                AgentMethod.UNKNOWN -> MissingMethodCallAgent(method = method)
                AgentMethod.STOP_TRACKING -> StopTrackingAgent()
                AgentMethod.SET_URL_MASKING_PATTERNS -> SetURLMaskingPatternsAgent()
                AgentMethod.SEND_TRANSACTION -> SendTransactionAgent()
                AgentMethod.SEND_TOUCH_EVENT -> SendTouchEventAgent()
                AgentMethod.SEND_SESSION_REPLAY_EVENT -> SendSessionReplayEventAgent()
                AgentMethod.SEND_SESSION_REPLAY_PROTO_DATA -> SendSessionReplayProtoDataListAgent()
                AgentMethod.SEND_SCREEN_NAME -> SendScreenNameAgent()
                AgentMethod.SEND_DYNAMIC_VAR -> SendDynamicVarAgent()
                AgentMethod.SEND_API_ERROR -> SendApiErrorAgent()
                AgentMethod.RESUME_TRACKING -> ResumeTrackingAgent()
                AgentMethod.GET_USER_ID -> GetUserIdAgent()
                AgentMethod.GET_CURRENT_SESSION_REPLAY_URL -> GetCurrentSessionReplayLinkAgent()
                AgentMethod.FORGET_ME -> ForgetMeAgent()
                AgentMethod.OPT_OUT -> OptOutAgent()
                AgentMethod.OPT_IN -> OptInAgent()
                AgentMethod.COLLECT_XPF_VERSION -> CollectXPFVersionAgent()
                AgentMethod.COLLECT_CUSTOM_TELEMETRY_DATA -> CollectCustomTelemetryAgent()
                AgentMethod.SEND_USER_IDENTIFIER -> SendUserIdentifierAgent()
                AgentMethod.GET_STORED_CONFIG -> GetStoredConfigAgent()
                AgentMethod.SEND_CRASH_REPORT -> SendCrashReportAgent()
                AgentMethod.WEB_VIEW_CREATE -> WebViewCreateAgent()
                AgentMethod.WEB_VIEW_DISPOSE -> WebViewDisposeAgent()
                AgentMethod.LISTEN_TO_SESSION_REPLAY_LINK_CHANGE -> OnSessionReplayLinkChangeAgent(channel = channel)
            }

        }
    }
}

enum class AgentMethod(val method: String) {
    UNKNOWN(method = "__unknown_method__"),
    STOP_TRACKING(method = "stopTracking"),
    SET_URL_MASKING_PATTERNS(method = "setURLMaskingPatterns"),
    SEND_TRANSACTION(method = "sendTransaction"),
    SEND_TOUCH_EVENT(method = "sendTouchEvent"),
    SEND_SESSION_REPLAY_EVENT(method = "sendSessionReplayEvent"),
    SEND_SESSION_REPLAY_PROTO_DATA(method = "sendSessionReplayProtoData"),
    SEND_SCREEN_NAME(method = "send"),
    SEND_DYNAMIC_VAR(method = "sendDynamicVar"),
    SEND_API_ERROR(method = "sendApiError"),
    RESUME_TRACKING(method = "resumeTracking"),
    GET_USER_ID(method = "getUserId"),
    GET_CURRENT_SESSION_REPLAY_URL(method = "currentSessionReplayUrl"),
    FORGET_ME(method = "forgetMe"),
    OPT_OUT(method = "optOut"),
    OPT_IN(method = "optIn"),
    COLLECT_XPF_VERSION(method = "collectXPFVersion"),
    COLLECT_CUSTOM_TELEMETRY_DATA(method = "collectCustomTelemetryData"),
    SEND_USER_IDENTIFIER(method = "sendUserIdentifier"),
    GET_STORED_CONFIG(method = "getInitialConfig"),
    SEND_CRASH_REPORT(method = "sendCrashReport"),
    WEB_VIEW_CREATE(method = "webViewCreate"),
    WEB_VIEW_DISPOSE(method = "webViewDispose"),
    LISTEN_TO_SESSION_REPLAY_LINK_CHANGE(method = "listenToSessionReplayLinkChange"),
}
