package com.contentsquare.flutter.infrastructure.core.platform_views

import android.webkit.WebView
import com.contentsquare.flutter.infrastructure.bridges.XpfInterfaceBridge
import com.contentsquare.flutter.infrastructure.core.method_call_executor.agents.OnWebViewCreateHandler
import com.contentsquare.flutter.infrastructure.core.method_call_executor.agents.OnWebViewDisposeHandler
import java.lang.ref.WeakReference
import java.util.WeakHashMap

/**
 * It is responsible to handle the registration of a WebView to the
 * Contentsquare SDK through our [XpfInterfaceBridge].
 *
 * It will handle events from the PlatformViews method channel that are reported
 * by the [PlatformViewsSpy] and events send from dart through the plugin MethodChannel
 * so that the WebView will be register and unregister at the expected moment.
 */
internal class WebViewRegistrationManager(
    private val viewFromIdProvider: ViewFromIdProvider,
    private val xpfInterfaceBridge: XpfInterfaceBridge = XpfInterfaceBridge.default()
): OnWebViewCreateHandler, OnWebViewFoundHandler, OnWebViewDisposeHandler {
    private val registeredWebViews = WeakHashMap<WebView, Int>()
    private val webViewsToRegister = HashMap<Int, WeakReference<WebView>>()
    private val viewIdsToRegister = HashMap<Int, Long>()
    private val scanners = HashMap<Int, WebViewScanner>()

    companion object {
        private var instance: WebViewRegistrationManager? = null

        /**
         * Get the instance of [WebViewRegistrationManager].
         */
        fun getInstance(): WebViewRegistrationManager? {
            return instance
        }

        /**
         * Initialize the instance of [WebViewRegistrationManager].
         */
        fun initInstance(flutterEngineProvider: FlutterEngineProvider): WebViewRegistrationManager {
            instance = WebViewRegistrationManager(
                viewFromIdProvider = ViewFromIdProvider(
                    flutterEngineProvider = flutterEngineProvider
                )
            )

            return instance as WebViewRegistrationManager
        }
    }

    private fun processWebViewRegistration(webView: WebView, webViewId: Long, viewId: Int) {
        if (!registeredWebViews.contains(webView)) {
            registeredWebViews[webView] = viewId
            xpfInterfaceBridge.registerWebView(webView, webViewId)
        }
        scanners[viewId]?.let { scanner ->
            scanner.stop()
        }
        scanners.remove(viewId)
    }

    private fun processWebViewUnRegistration(webView: WebView) {
        if (registeredWebViews.contains(webView)) {
            xpfInterfaceBridge.unregisterWebView(webView)
            registeredWebViews.remove(webView)
        }
    }

    override fun onWebViewCreate(webViewId: Long, viewId: Int) {
        val view = viewFromIdProvider.getView(viewId)
        if (view is WebView) {
            processWebViewRegistration(view, webViewId, viewId)
        } else {
            viewIdsToRegister[viewId] = webViewId
            val scanner = WebViewScanner(
                viewId = viewId,
                viewFromIdProvider = viewFromIdProvider,
                handler = this
            )
            scanner.start()
            scanners[viewId] = scanner
        }
    }

    override fun onWebViewDispose(webViewId: Long, viewId: Int) {
        scanners[viewId]?.let { scanner ->
            scanner.stop()
        }
        registeredWebViews.forEach { entry ->
            if (entry.value == viewId && entry.key != null) {
                processWebViewUnRegistration(entry.key)
            }
        }

        webViewsToRegister.remove(viewId)
        viewIdsToRegister.remove(viewId)
    }

    override fun onWebViewFound(webView: WebView, viewId: Int) {
        viewIdsToRegister[viewId]?.let { webViewId ->
            processWebViewRegistration(webView, webViewId, viewId)
        }
    }
}