package com.contentsquare.flutter.infrastructure.core.platform_views

import android.os.Handler
import android.os.Looper
import android.webkit.WebView
import java.util.Timer
import kotlin.concurrent.schedule

/**
 * It allows to execute at [intervalScan] a scan to look for a [WebView]
 * that was added to the tree through the Flutter PlatformViews API.
 *
 * It will use periodically [viewFromIdProvider] to try to get [WebView]
 * registered by [viewId] with a maximum delay of [timeoutScan] and pass it
 * to the [handler].
 */
internal class WebViewScanner(
    private val viewFromIdProvider: ViewFromIdProvider,
    private val viewId: Int,
    private val handler: OnWebViewFoundHandler,
    private val intervalScanMs: Long = 250,
    private val timeoutScanMs: Long = 5000,
    private val looperHandler: Handler = Handler(Looper.getMainLooper())
) {
    private var isScanning = false
    private var isStopped = false
    private var scanIntervalTimer: Timer? = null
    private var scanTimeoutTimer: Timer? = null

    /**
     * Whether the scan process is running.
     */
    val isWorking: Boolean
        get() = scanIntervalTimer != null

    /**
     * Start the scanning process.
     *
     * If there was one already running then it will be
     * stopped before the new one starts.
     */
    fun start() {
        if (isWorking) stop()

        scanIntervalTimer = Timer().apply {
            schedule(0, intervalScanMs) {
                looperHandler.post {
                    performScan()
                }
            }
        }

        scanTimeoutTimer = Timer().apply {
            schedule(timeoutScanMs) {
                looperHandler.post {
                    stop()
                }
            }
        }
    }

    /**
     * Stop the scanning process.
     */
    fun stop() {
        scanIntervalTimer?.cancel()
        scanIntervalTimer = null
        scanTimeoutTimer?.cancel()
        scanTimeoutTimer = null
        isScanning = false
        isStopped = true
    }

    private fun performScan() {
        if (isScanning || isStopped) return

        isScanning = true
        val webView = viewFromIdProvider.getView(viewId)
        if (webView is WebView) {
            handler.onWebViewFound(webView, viewId)
            stop()
        }
        isScanning = false
    }
}

/**
 * Handler called when a [WebView] is found by the [WebViewScanner].
 */
internal interface OnWebViewFoundHandler {
    /**
     * Handles the WebView found.
     */
    fun onWebViewFound(webView: WebView, viewId: Int)
}
