/// Allows to listen to session replay link changes from Contentsquare API
protocol OnSessionReplayLinkChangeProxy {
    
    /// Listens to the session replay link changes from Contentsquare API
    func onSessionReplayLinkChange() throws -> Void
}

/// Factory creator class for [OnSessionReplayLinkChangeProxy]
struct OnSessionReplayLinkChangeProxyFactory {
    
    /// Creates a new instance of the default implementation of [OnSessionReplayLinkChangeProxy].
    ///
    /// - Parameter contentsquareSDK: The `ContentsquareSDKBridge` used to access contentsquare API.
    /// - Parameter flutterMethodChannel: the `FlutterMethodChannel` used to access channel's methods
    static func makeDefault(contentsquareSDK: ContentsquareSDKBridge = ContentsquareSDKBridgeFactory.getInstance(), flutterMethodChannel: FlutterMethodChannel) -> OnSessionReplayLinkChangeProxy { OnSessionReplayLinkChangeProxyImpl(contentsquareSDK: contentsquareSDK, flutterMethodChannel: flutterMethodChannel)
    }
}


// An implementation of [OnSessionReplayLinkChangeProxy]
class OnSessionReplayLinkChangeProxyImpl:
    OnSessionReplayLinkChangeProxy {
    
    private let _contentsquareSDK: ContentsquareSDKBridge
    private let _channel: FlutterMethodChannel
    
    /// Creates a new instance of `OnSessionReplayLinkChangeProxyImpl`
    ///
    /// - Parameter contentsquareSDK: The `ContentsquareSDKBridge` used to access contentsquare API.
    /// - Parameter flutterMethodChannel: the `FlutterMethodChannel` used to access channel's methods
    internal init(contentsquareSDK: ContentsquareSDKBridge, flutterMethodChannel: FlutterMethodChannel){
        self._contentsquareSDK = contentsquareSDK
        self._channel = flutterMethodChannel
    }
    
    func onSessionReplayLinkChange() throws {
        try _contentsquareSDK.onSessionReplayLinkChange{ link in
            
            guard let link = link else { return }

            let arguments: [String: Any] = [FlutterFeaturesKey.sessionReplayUrl.rawValue: link.absoluteString as String]
            self._channel.invokeMethod(FlutterChannelMethod.notifySessionReplayUrlChange.rawValue, arguments: arguments)
            
        }
    }
}

