import Flutter
import WebKit
import UIKit
import ContentsquareModule

// This is the protocol that the listeners passed by external bridges have to implement.
@objc
protocol ExternalBridgeInterface {
    @objc
    func takeSnapshot(parameters: [String: Any])
    
    @objc
    func enableSessionReplay(parameters: [String: Any])

    @objc
    func enableAPIErrors(parameters: [String: Any])

    @objc
    func enableCrashReporter(parameters: [String: Any])

    @objc
    func updateFeatureFlags(_: [[String: Any]])
    
    @objc
    func notifyCSInAppEnabled(parameters: [String: Any])
    
    @objc
    func shouldHandleWebView(_: AnyObject) -> NSNumber
    
    @objc
    func shouldHandleExternalBridge(_: AnyObject) -> NSNumber

}

private typealias TakeSnapshotHandler = (_ rootPath: String, _ completion: @escaping (String?) -> Void) -> Void
private typealias EnableSessionReplayHandler = (_ enable: Bool, _ containerRecordingId: Int?) -> Void
private typealias EnableAPIErrorHandler = (_ enable: Bool) -> Void
private typealias EnableCrashReporterHandler = (_ enable: Bool) -> Void
private typealias UpdateFeatureFlagsHandler = (_ featureFlags: [[String: Any]]) -> Void
private typealias EnableCsInAppHandler = (_ enable: Bool) -> Void
private typealias WebViewSupportsByNativeCheckerProvider = () -> WebViewSupportsByNativeChecker?

private class ExternalBridgeImpl: ExternalBridgeInterface {
    private let _targetViewPathCreator = TargetViewPathCreator()
    
    private var _webViewSupportsByNativeCheckerProvider: WebViewSupportsByNativeCheckerProvider
    private var _takeSnapshotHandler: TakeSnapshotHandler
    private var _enableSessionReplayHandler: EnableSessionReplayHandler
    private var _enableAPIErrorHandler: EnableAPIErrorHandler
    private var _enableCrashReporterHandler: EnableCrashReporterHandler
    private var _updateFeatureFlagsHandler: UpdateFeatureFlagsHandler
    private var _pluginState : SwiftContentsquarePluginState
    private var _enableCsInAppHandler: EnableCsInAppHandler

    
    init(takeSnapshotHandler: @escaping TakeSnapshotHandler, enableSessionReplayHandler: @escaping EnableSessionReplayHandler, enableAPIErrorHandler: @escaping EnableAPIErrorHandler, enableCrashReporterHandler: @escaping EnableCrashReporterHandler, updateFeatureFlagsHandler: @escaping UpdateFeatureFlagsHandler,
         enableCsInAppHandler: @escaping EnableCsInAppHandler, webViewSupportsByNativeCheckerProvider: @escaping WebViewSupportsByNativeCheckerProvider) {
        _takeSnapshotHandler = takeSnapshotHandler
        _enableSessionReplayHandler = enableSessionReplayHandler
        _enableAPIErrorHandler = enableAPIErrorHandler
        _enableCrashReporterHandler = enableCrashReporterHandler
        _updateFeatureFlagsHandler = updateFeatureFlagsHandler
        _enableCsInAppHandler = enableCsInAppHandler
        _pluginState = SwiftContentsquarePluginState.shared
        _webViewSupportsByNativeCheckerProvider = webViewSupportsByNativeCheckerProvider
    }
    
    @objc
    func takeSnapshot(parameters: [String: Any]) {        
        if let rootPath = parameters[TakeSnapshotParam.rootPath.rawValue] as? String,
           let completion = parameters[TakeSnapshotParam.completion.rawValue] as? (String?) -> Void {
            let targetViewPath = _targetViewPathCreator.create(rootPath: rootPath)
            _takeSnapshotHandler(targetViewPath, completion)
        }
    }
    
    @objc
    func enableSessionReplay(parameters: [String: Any]) {
        if let enable = parameters[EnableSessionReplayParam.enable.rawValue] as? Bool {
            let containerRecordingId = parameters[EnableSessionReplayParam.containerRecordingId.rawValue] as? Int

            _enableSessionReplayHandler(enable, containerRecordingId)
            _pluginState.setSessionReplayState(StateObject(value: enable))
            
            if (containerRecordingId != nil) {
                _pluginState.setsessionReplayContainerRecordingId(StateObject(value: containerRecordingId!))
            }
        }
    }

    @objc
    func enableAPIErrors(parameters: [String : Any]) {
        if let enable = parameters[EnableAPIErrorParam.enable.rawValue] as? Bool {
            _enableAPIErrorHandler(enable)
            _pluginState.setApiErrorsState(StateObject(value: enable))
        }
    }
    
    @objc
    func enableCrashReporter(parameters: [String : Any]) {
        if let enable = parameters[EnableCrashReporterParam.enable.rawValue] as? Bool {
            _enableCrashReporterHandler(enable)
            _pluginState.setCrashReporterState(StateObject(value: enable))
        }
    }

    @objc
    func updateFeatureFlags(_ featureFlags: [[String: Any]]) {
        _updateFeatureFlagsHandler(featureFlags);
        _pluginState.setFeatureFlagsState(StateObject(value: featureFlags))
    }
    
    @objc
    func notifyCSInAppEnabled(parameters: [String : Any]) {
        if let enable = parameters[NotifyCSInAppEnabledParam.enable.rawValue] as? Bool {
            _enableCsInAppHandler(enable)
            _pluginState.setCsInAppState(StateObject(value: enable))
        }

    }
    
    @objc
    func shouldHandleWebView(_ view: AnyObject) -> NSNumber {
        guard let webView = view as? WKWebView else {
            return NSNumber(booleanLiteral: false)
        }
        
        if let checker = _webViewSupportsByNativeCheckerProvider() {
            if checker.shouldNativeSupports(webView) {
                return NSNumber(booleanLiteral: false)
            }
        }
        
        return NSNumber(booleanLiteral: true)
    }
    
    @objc
    func shouldHandleExternalBridge(_ view: AnyObject) -> NSNumber {
        if let uiView = view as? UIView, let superView = uiView.superview {
            /// We need to handle the view that is used by Flutter to display the UI of the application.
            /// This UI is displayed inside the UIScrollView that is the subview of the FlutterView
            /// But inside this FlutterView we also have others UIScrollView for semantics purpose and
            /// those ones we don't handled them.
            ///
            /// So we get the view that is a subview and a UIScrollView but not a FlutterSemanticsScrollView
            if superView.className == "FlutterView" {
                return NSNumber(booleanLiteral: uiView is UIScrollView && uiView.className != "FlutterSemanticsScrollView")
            }
        }
        
        return NSNumber(booleanLiteral: false)
    }
}

// Expected elements takeSnapshot's parameters dictionary:
enum TakeSnapshotParam: String {
    
    // path of the rootView of the bridge on the native side. To be used to create TargetViewPaths when generating the snapshot
    case rootPath // as String - mandatory
    
    // completion block to be used in order to return the json of the generated snapshot asynchronoulsy
    case completion // as (json:String?) -> Void - mandatory
}

enum EnableSessionReplayParam: String {
    
    // flag, that represents the new status of the session recording.
    case enable // as bool - mandatory
    case containerRecordingId // as Int - mandatory
}

enum EnableAPIErrorParam: String {

    // flag, that represents the new status of the API Error.
    case enable // as bool - mandatory
}

enum EnableCrashReporterParam: String {

    // flag, that represents the new status of the crash reporter.
    case enable // as bool - mandatory
}

enum UpdateFeatureFlagsParam: String {

    // list of features flags
    case featureFlags // as [[String: Any]] - mandatory
}

// Expected elements notifyCsInAppEnabled's parameters dictionary:
enum NotifyCSInAppEnabledParam: String {

    // flag, that represents the new status of the CSInApp.
    case enable // as bool - mandatory
}

enum FlutterChannelMethod: String {
    case takeSnapshot = "requestFlutterSnapshot"
    case enableSR = "requestEnableSessionReplay"
    case enableAPIError = "requestEnableAPIError"
    case enableCrashReporter = "requestEnableCrashReporter"
    case updateFeatureFlags = "updateFeatureFlags"
    case enableCsInApp = "requestEnableCsInApp"
    case notifySessionReplayUrlChange = "requestSessionReplayUrlChange"
}

enum FlutterFeaturesKey : String {
    case enableSR = "enableSR"
    case containerRecordingId = "containerRecordingId"
    case enableApiErrors = "enableApiErrors"
    case enableCrashReporter = "enableCrashReporter"
    case featureFlags = "featureFlags"
    case enableCsInApp = "enableCsInApp"
    case sessionReplayUrl = "sessionReplayUrl"
    case appInfo = "appInfo"
}

func createExternalBridge(channel: FlutterMethodChannel, webViewSupportsByNativeChecker: WebViewSupportsByNativeChecker?) -> ExternalBridgeInterface {
    let flutterTag = "|flutter|"
    
    func snapshotHandler(rootPath: String, completion: @escaping (String?) -> Void) {
        let arguments: [String: Any] = [TakeSnapshotParam.rootPath.rawValue: rootPath + flutterTag]
        channel.invokeMethod(FlutterChannelMethod.takeSnapshot.rawValue, arguments: arguments) { result in
            var json = result as? String
            completion(json)
        }
    }
    
    func enableSRHandler(enable: Bool, _ containerRecordingId: Int?) {
        let arguments: [String: Any?] = [FlutterFeaturesKey.enableSR.rawValue: enable,
                                        FlutterFeaturesKey.containerRecordingId.rawValue: containerRecordingId]
        channel.invokeMethod(FlutterChannelMethod.enableSR.rawValue, arguments: arguments)
    }

    func enableAPIErrorHandler(enable: Bool) {
        let arguments: [String: Any] = [FlutterFeaturesKey.enableApiErrors.rawValue: enable]
        channel.invokeMethod(FlutterChannelMethod.enableAPIError.rawValue, arguments: arguments)
    }
    
    func enableCrashReporterHandler(enable: Bool) {
        let arguments: [String: Any] = [FlutterFeaturesKey.enableCrashReporter.rawValue: enable]
        channel.invokeMethod(FlutterChannelMethod.enableCrashReporter.rawValue, arguments: arguments)
    }

    func updateFeatureFlagsHandler(featureFlags: [[String: Any]]) {
        let arguments: [String: Any] = [UpdateFeatureFlagsParam.featureFlags.rawValue: featureFlags]
        channel.invokeMethod(FlutterChannelMethod.updateFeatureFlags.rawValue, arguments: arguments)
    }
    
    func enableCsInAppHandler(enable: Bool) {
        let arguments: [String: Any] = [FlutterFeaturesKey.enableCsInApp.rawValue: enable]
        channel.invokeMethod(FlutterChannelMethod.enableCsInApp.rawValue, arguments: arguments)
    }
    
    func webViewSupportsByNativeCheckerProvider() -> WebViewSupportsByNativeChecker? {
        return webViewSupportsByNativeChecker ?? WebViewRegistrationManager.shared
    }
    
    return ExternalBridgeImpl(
        takeSnapshotHandler: snapshotHandler,
        enableSessionReplayHandler: enableSRHandler,
        enableAPIErrorHandler: enableAPIErrorHandler,
        enableCrashReporterHandler: enableCrashReporterHandler,
        updateFeatureFlagsHandler: updateFeatureFlagsHandler,
        enableCsInAppHandler: enableCsInAppHandler,
        webViewSupportsByNativeCheckerProvider: webViewSupportsByNativeCheckerProvider
    )
}

func registerExternalBridge(channel: FlutterMethodChannel) {
    
    let interface = createExternalBridge(channel: channel, webViewSupportsByNativeChecker: nil)
    
    Contentsquare.perform(
        NSSelectorFromString("_registerExternalBridgeWithParameters:"),
        with: [
            "interface": interface,
            "type": 1
        ]
    )
}

public class SwiftContentsquarePlugin: NSObject, FlutterPlugin {
    private var executor: MethodCallExecutor
    static private var _channel: FlutterMethodChannel!
    
    init(executor: MethodCallExecutor = MethodCallExecutor()) {
        self.executor = executor
    }
    
    public static func register(with registrar: FlutterPluginRegistrar) {
        if (SwiftContentsquarePlugin._channel != nil) {
            NSLog(
            """
            CSLIB ⚠️ WARNING: SwiftContentsquarePlugin was registered twice. Please refer to the Contentsquare documentation's troubleshooting section for more information about what to do: https://docs.contentsquare.com/flutter/#troubleshooting
            """
            );
            return;
        }
        let channel = FlutterMethodChannel(name: "contentsquare", binaryMessenger: registrar.messenger())
        SwiftContentsquarePlugin._channel = channel
        
        let instance = SwiftContentsquarePlugin()
        registrar.addMethodCallDelegate(instance, channel: channel)
        registerExternalBridge(channel: channel)
        SwiftContentsquarePluginState.shared.setAppInfoState(StateObject(value: AppInfoProvider.get()))
        registrar.addApplicationDelegate(instance)
    }
    
    public func handle(_ call: FlutterMethodCall, result: @escaping FlutterResult) {
        let agent: MethodCallAgent = MethodCallAgentFactory.makeInstance(method: call.method, channel: SwiftContentsquarePlugin._channel)
        let executionData = ExecutionData(call: call)
        
        self.executor.execute(agent: agent, data: executionData, result: result)
    }
    
    public func application(_ application: UIApplication, didFinishLaunchingWithOptions launchOptions: [AnyHashable : Any] = [:]) -> Bool {
        WebViewRegistrationManager.initFrom(application: application)
        
        return true
    }
    
    public func applicationWillTerminate(_ application: UIApplication) {
        WebViewRegistrationManager.shared?.dispose()
    }
}

extension UIView {
    
    func allSubviews(where predicate: (UIView) -> Bool) -> [UIView] {
        let views: [UIView] = predicate(self) ? [self] : []
        return views + subviews.flatMap { $0.allSubviews(where: predicate) }
    }
    
    var className: String {
        String(describing: type(of: self))
    }
}
